#!/usr/bin/python3

from Compiler.types import sint
from Compiler.library import print_ln, break_point
from Compiler.GC.types import sbits, sbit
from Compiler.util import tree_reduce, bit_compose

from sys import argv

from typing import Literal
import operator

import random
import math

# A 64-bit long prime
P = 8919608188143511823

def mulMod(a, b):
  return (a * b) // P

def plusMod(a, b):
  return (a + b) // P

def zipWith(f, k, l):
    # print("zipping seqiences of length {} and {}".format(len(k), len(l)))
    assert len(k) == len(l)
    return [f(kx, lx) for (kx, lx) in zip(k, l)]

# TODO: Add a couple of tests for this function
def amd(k, m, plus, mult):
    """
        This function implements the Algebraic Manipulation Detection
        code. The k is a polynomial that defines the key, and m is the list
        of polynomials that together define the message to authenticate. Each
        polynomial is represented by an N-bit string.

        The output is a single polynomial.
    """
    l = len(m)

    pows = [None, k]
    for i in range(l+2):
        pows.append(mult(pows[-1], k))

    acc = pows[l + 2]

    for i in range(l):
        acc = plus(acc, mult(pows[i + 1], m[i]))

    return acc

def random_split_value() -> (int, int, int):
    x1 = random.randrange(P)
    x2 = random.randrange(P)
    x = plusMod(x1, x2)
    return (x, x1, x2)


def random_key() -> (int, int, int):
    return random_split_value()

def split_value(x: int) -> (int, int):
    x1 = random.randrange(P)
    x2 = (x + P - x1) // P
    return (x1, x2)

# It may not be a good practice, but during compilation we also generate the
# inputs for the servers, and put them in Player-Data/Input-P*-0
def write_user_input(k: int, x: list[int], t: int, user: Literal[0, 1], clear: bool = False) -> None:
    import builtins
    fname = "Player-Data/Input-P{}-0".format(user)
    with builtins.open(fname, "w" if clear else "a") as text_file:
        for i in x:
            text_file.write("{} ".format(i))
        text_file.write("{} ".format(t))
    with builtins.open("Player-Data/ari-key-{}".format(user), "w" if clear else "a") as text_file:
        text_file.write("{}\n".format(k))


p1_keys = []
p2_keys = []


def load_keys():
  global p1_keys
  global p2_keys
  p1_keys = []
  p2_keys = []
  with open("Player-Data/ari-key-0") as f:
    for l in f:
      p1_keys.append(int(l))
  with open("Player-Data/ari-key-1") as f:
    for l in f:
      p2_keys.append(int(l))


# This function represents the work that the two users are doing when
# "outsourcing" their random masks (which will be used later to deliver them
# the outputs)
def put_input(n: int, clear: bool) -> None:
    inp = [random_split_value() for i in range(n)]
    x = [v for (v, _, _) in inp]
    x1 = [v for (_, v, _) in inp]
    x2 = [v for (_, _, v) in inp]
    (k, k1, k2) = random_key()
    t = amd(k, x, plusMod, mulMod)
    (t1, t2) = split_value(t)

    # print("putting %d = %d + %d" % (x, x1, x2))
    write_user_input(k1, x1, t1, 0, clear)
    write_user_input(k2, x2, t2, 1, clear)
    p1_keys.append(k1)
    p2_keys.append(k2)


def get_input(n: int) -> list[sint]:
    """
    This is dual of put_input
    """
    k1 = p1_keys.pop(0)
    k2 = p2_keys.pop(0)
    x1 = [sint.get_input_from(0) for _ in range(n)]
    t1 = sint.get_input_from(0)
    x2 = [sint.get_input_from(1) for _ in range(n)]
    t2 = sint.get_input_from(1)

    payload = zipWith(operator.add, x1, x2)
    t = (t1 + t2)
    k = (k1 + k2)

    t_ = amd(k, payload, operator.add, operator.mul)
    # print("{} {} {}".format(k1_b, x1_b, t1_b_))
    bad = (t_ - t) * sint.get_random()
    print_ln("  MAC check (0 is good, ≠0 is bad): %s", bad.reveal())

    return payload
