#!/usr/bin/python3

from Compiler.types import sint
from Compiler.library import print_ln, break_point
from Compiler.GC.types import sbits, sbit
from Compiler.util import tree_reduce, bit_compose

from sys import argv

from typing import Literal

import random
import math

# for xor, and
import operator

def shiftLeft(v, k):
    r = v.copy()
    for i in range(k):
        r.pop(0)
    return r

def shiftRightAndPad(v, k, n, zero):
    # print("shiftRightAndPad: {} {} {} {}".format(v, k, n, zero))
    r = v.copy()
    for i in range(k):
        r.insert(0, zero)
    for i in range(n - len(r)):
        r.append(zero)
    assert len(r) - 1 <= n
    # print("shiftRightAndPad r = {}".format(r))
    return r

def foldl1(f, l):
    res = l.pop(0)
    while len(l) != 0:
        res = f(res, l.pop(0))
    return res

def zipWith(f, k, l):
    # print("zipping seqiences of length {} and {}".format(len(k), len(l)))
    assert len(k) == len(l)
    return [f(kx, lx) for (kx, lx) in zip(k, l)]

def matrixApply(A, v, plus, mult):
    assert len(A[0]) == len(v)
    ret = [foldl1(plus, zipWith(mult, a, v)) for a in A]
    assert len(ret) == len(A)
    return ret

def matrixTrans(A):
    n = len(A)
    m = len(A[0])
    return [[A[i][j] for i in range(n)] for j in range(m)]

def matrixProd(A, B, plus, mult):
    assert len(A[0]) == len(B)
    return matrixTrans([matrixApply(A, v, plus, mult) for v in matrixTrans(B)])

# Returns a matrix A such that (x ↦ Ax) = (x ↦ v · x), where v · x is the
# product of two polynomials modulo 2 of degree n-1 both. A has the dimension
# n×(2n-1)
#
# TODO: Add some tests for this function
def productMatrix(v, plus, mult, zero, one):
    n = len(v)
    def cond_unit(i, x):
        return [x if j == i else zero for j in range(n)]
    res = [foldl1(lambda u, v: zipWith(plus, u, v), [cond_unit(i, v[d-i]) for i in range(max(0, d + 1 - n), min(d + 1, n))]) for d in range(2 * n - 1)]
    # print("product matrix is {}×{}".format(len(res), len(res[0])))
    assert len(res) == 2 * n - 1
    assert len(res[0]) == n
    return res

# Returns matrix B such that (x ↦ Bx) = (x ↦ x mod p) where x and p are
# polynomials mod 2 with the corresponding degrees 2n-2 (the result of
# multiplying two polynomials of degrees n-1) and n. B has dimension n×(2n - 1)
#
# TODO: Add some test cases for this function
def modulusMatrix(p, plus, mult, zero, one):
    n = len(p) - 1
    # print("n = %d" % n)
    def unit(i):
        return [one if j == i else zero for j in range(2 * n - 1)]
    # def cnst(b):
    #     return [zero for j in range(2 * n - 1)] + [b]
    def mult_(a, v):
        return [mult(a, x) for x in v]
    res = [unit(i) for i in range(2 * n - 1)]
    # print("res before = {}".format(res))
    for deg in reversed(range(n, 2 * n - 1)):
        # deg is the degree we're eliminating
        pConst = shiftRightAndPad(p, deg - (n), 2 * n - 1, zero)
        pVec = [mult_(p_i, res[deg]) for p_i in pConst]
        # print("pVec = {}".format(pVec))
        res = [zipWith(plus, l, r) for (l, r) in zip(pVec, res)]
    # print("res after = {}".format(res))
    ret = res[0:n]
    # print("modulus matrix is {}×{}, {}".format(len(ret), len(ret[0]), ret))
    assert len(ret[0]) == 2 * n - 1
    return ret


# The matrix that implements (x ↦ b · x mod p) as a linear operator. Has dimension n×n
def prodModMatrix(b, p, plus, mult, zero, one):
    return matrixProd(modulusMatrix(p, plus, mult, zero, one), productMatrix(b, plus, mult, zero, one), plus, mult)

def multMod(u, v, p, plus, mult, zero, one):
    return matrixApply(prodModMatrix(u, p, plus, mult, zero, one), v, plus, mult)

def addMod(u, v, plus):
    return zipWith(plus, u, v)

def squareMod(x, p, plus, mult, zero, one):
    return multMod(x, x, p, plus, mult, zero, one)

def powMod(v, t, p, plus, mult, zero, one):
    if t == 1:
        return v
    elif t % 2 == 0:
        acc = powMod(v, t // 2, p, plus, mult, zero, one)
        return squareMod(acc, p, plus, mult, zero, one)
    else:
        acc = powMod(v, t - 1, p, plus, mult, zero, one)
        return multMod(acc, v, p, plus, mult, zero, one)

assert [True, True, False] == matrixApply(modulusMatrix([True, False, False, True], operator.xor, operator.and_, False, True), [True, False, False, False, True], operator.xor, operator.and_)
assert [True, True, False] == squareMod([True, False, True], [True, False, False, True], operator.xor, operator.and_, False, True)

# The dimention of the polynomials that we use. In other words, we are working
# in F₂[x]/p where p has the degree n. This value is also our statistical
# security parameter
N = 41
P = [True if i in [0, 3, 41] else False for i in range(42)]
assert len(P) == 42

# TODO: Add a couple of tests for this function
def amd(k, m, plus, mult, zero, one):
    """
        This function implements the Algebraic Manipulation Detection
        code. The k is a polynomial that defines the key, and m is the list
        of polynomials that together define the message to authenticate. Each
        polynomial is represented by an N-bit string.

        The output is a single polynomial.
    """
    l = len(m)

    pows = [None, k]
    kProd = prodModMatrix(k, P, plus, mult, False, True) # The hardcoded True and False here are important
    for i in range(l+2):
        pows.append(matrixApply(kProd, pows[-1], plus, mult))

    acc = pows[l + 2]

    for i in range(l):
        acc = addMod(acc, multMod(pows[i + 1], m[i], P, plus, mult, False, True), plus)

    return acc

# print("{}".format(amd([True if i <= 10 else False for i in range(N)], [[True if i <= 5 else False for i in range(N)]], operator.xor, operator.and_, False, True)))

# The next few functions are just utilities for compilation


def from_bits(n):
    res = 0
    for (i, x) in enumerate(n):
        res += 2 ** i * x
    l = len(n)
    if res < 2 ** (l - 1):
        return res
    else:
        return res - 2 ** l


def reverseIWishIUsedHaskell(l):
    return list(reversed(l))


def to_bits(x, l):
    assert x >= -(2 ** (l-1))
    assert x < 2 ** (l-1)
    if x < 0:
        x = x + 2 ** l
    return [True if (x & (1 << i)) >> i == 1 else False for i in range(l)]


assert from_bits([0, 1, 0, 0, 1, 0]) == 18
assert from_bits(reverseIWishIUsedHaskell([1, 0, 0, 0, 0, 0, 0, 0])) == -128
assert from_bits(reverseIWishIUsedHaskell([1, 1, 1, 1, 1, 1, 1, 1])) == -1
assert to_bits(18, 6) == [0, 1, 0, 0, 1, 0]
for i in [0, 1, -1, -10, 5, -8]:
    assert from_bits(to_bits(i, 5)) == i


FieldT = sbits.get_type(N + 1)


def random_split_value(l) -> (int, int, int):
    x1 = random.getrandbits(l) - 2 ** (l - 1)
    x2 = random.getrandbits(l) - 2 ** (l - 1)
    x = x1 ^ x2
    return (x, x1, x2)


def random_key() -> (int, int, int):
    return random_split_value(N)

def split_value(x: int, l: int) -> (int, int):
    x1 = random.getrandbits(l) - 2 ** (l - 1)
    x2 = x ^ x1
    return (x1, x2)

def tag(n: int, x, k):
    """Pick a random key and authenticate the given value with it. Return both
    key and tag."""
    xNum = x
    # print("tag({}, {}, {})\n".format(n, x, k))
    blocks = math.ceil(n / N)
    x_ = to_bits(x, n)
    x = chop_input(x_)
    k = to_bits(k, N)
    t = from_bits(amd(k, x, operator.xor, operator.and_, False, True))
    # print("picked AMD({}, {} = {}) = {}".format(from_bits(k), xNum, [from_bits(i) for i in x], t))
    return t

# It may not be a good practice, but during compilation we also generate the
# inputs for the servers, and put them in Player-Data/Input-P*-0
def write_user_input(k: int, x: int, t: int, user: Literal[0, 1], clear: bool = False) -> None:
    import builtins
    fname = "Player-Data/Input-P{}-0".format(user)
    with builtins.open(fname, "w" if clear else "a") as text_file:
        text_file.write("{} {} ".format(x, t))
    with builtins.open("Player-Data/bin-key-{}".format(user), "w" if clear else "a") as text_file:
        text_file.write("{}\n".format(k))


p1_keys = []
p2_keys = []

def load_keys():
  global p1_keys
  global p2_keys
  p1_keys = []
  p2_keys = []
  with open("Player-Data/bin-key-0") as f:
    for l in f:
      p1_keys.append(int(l))
  with open("Player-Data/bin-key-1") as f:
    for l in f:
      p2_keys.append(int(l))


# This function represents the work that the two users are doing when
# "outsourcing" their random masks (which will be used later to deliver them
# the outputs)
def put_input(n: int, clear: bool) -> None:
    (x, x1, x2) = random_split_value(n)
    (k, k1, k2) = random_key()
    t = tag(n, x, k)
    (t1, t2) = split_value(t, N)

    # print("putting %d = %d + %d" % (x, x1, x2))
    write_user_input(k1, x1, t1, 0, clear)
    write_user_input(k2, x2, t2, 1, clear)
    p1_keys.append(k1)
    p2_keys.append(k2)


def chop_input(l):
    blocks = math.ceil(len(l) / N)
    l = l + [False for i in range(blocks * N - len(l))]
    return [l[i * N: (i+1) * N] for i in range(blocks)]

def get_input(n: int) -> list[sbit]:
    """
    This is dual of put_input
    """
    T = sbits.get_type(n + 1) # TODO: Do I need the +1 here?

    k1 = p1_keys.pop(0)
    k2 = p2_keys.pop(0)
    x1 = T.get_input_from(0)
    t1 = FieldT.get_input_from(0)
    x2 = T.get_input_from(1)
    t2 = FieldT.get_input_from(1)

    x1_b = x1.bit_decompose()
    x2_b = x2.bit_decompose()
    t1_b = t1.bit_decompose()
    t2_b = t2.bit_decompose()
    for iWishPythonWasABetterLanguage in [x1_b, x2_b, t1_b, t2_b]:
        iWishPythonWasABetterLanguage.pop()

    payload = zipWith(operator.xor, x1_b, x2_b)
    t_b     = zipWith(operator.xor, t1_b, t2_b)
    k_b = to_bits(k1 ^ k2, N)

    or_ = lambda x, y: 1 ^ ((1 ^ x) & (1 ^ y))
    zero = sbit(0)
    one = sbit(1)

    t_b_ = amd(k_b, chop_input(payload), operator.xor, operator.and_, zero, one)
    # print("{} {} {}".format(k1_b, x1_b, t1_b_))
    bad = tree_reduce(or_, zipWith(operator.xor, t_b, t_b_))
    print_ln("  MAC check (0 is good, 1 is bad): %s", bad.reveal())

    return payload
